#include "common/compat.h"

#include "gdb-serial.h"
#include "pinouts/pinouts.h"

GDB_Serial::GDB_Serial(Interconnect* ic)
 : Component(ic),
   uart  (*addMasterPort("uart",   uint8_pinout)),
   echo  (*addSlavePort ("echo",   uint8_pinout)),
   packet(*addSlavePort ("packet", shift_pinout)),
   tail(0) {
}

void GDB_Serial::recv(RecvQueue::const_iterator i, RecvQueue::const_iterator end) {
  static const char hex[] = "0123456789abcdef";
  
  for (; i != end; ++i) {
    if (i->slave == &echo) {
      uart(i->bits);
    } else {
      uart(Bits(static_cast<uint8_t>('$')));
      uint8_t chksum = 0;

      const Bits& bits = i->bits;
      Bits::size_type len = bits.size() / 8;
      for (Bits::size_type j = 0; j < len; ++j) {
        uint8_t b = bits.get_uint8(j);
        uart(Bits(static_cast<uint8_t>(hex[b >> 4])));
        uart(Bits(static_cast<uint8_t>(hex[b & 0xff])));
        chksum += b;
      }
      uart(Bits(static_cast<uint8_t>('#')));
      uart(Bits(static_cast<uint8_t>(hex[chksum >> 4])));
      uart(Bits(static_cast<uint8_t>(hex[chksum & 0xff])));
    }
  }
}

SendQueue::const_iterator GDB_Serial::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->master == &reset) {
      tail = 0;
      buffer.clear();
      continue;
    }
    
    char c = static_cast<uint8_t>(i->bits);
    buffer.push_back(c);
    
    if (tail > 0) {
      if (++tail == 3) {
        runCommand();
        tail = 0;
      }
    } else {
      if (c == '#') tail = 1;
    }
  }
  
  return i;
}

static int hexDigit(char c) {
  if (c >= '0' && c <= '9')
    return c-'0';
  else if (c >= 'a' && c <= 'f')
    return c-'a'+10;
  else if (c >= 'A' && c <= 'F')
    return c-'A'+10;
  else
    return -1;
}

void GDB_Serial::runCommand() {
  /* Checksum values */
  int hex1 = hexDigit(buffer[buffer.size()-2]);
  int hex2 = hexDigit(buffer[buffer.size()-1]);
  
  /* Find the preceding '$' */
  std::string::size_type dollar = buffer.rfind('$');
  if (dollar == std::string::npos || hex1 == -1 || hex2 == -1) {
    buffer.clear();
    echo(Bits(static_cast<uint8_t>('-')));
    return;
  }
  
  std::string body(buffer, dollar, buffer.size()-dollar-3);
  buffer.clear();
  
  uint8_t sum = 0;
  for (std::string::size_type i = 0; i < body.size(); ++i)
    sum += body[i];
  
  if (sum == ((hex1<<4) | hex2)) {
    echo(Bits(static_cast<uint8_t>('+')));
    packet(Bits(body));
  } else {
    echo(Bits(static_cast<uint8_t>('-')));
  }
}
