#include "common/compat.h"

#include "pinouts/pinouts.h"
#include "lm32-read.h"

#include <assert.h>

LM32Read::LM32Read(Interconnect* ic, bool s)
 : LM32Base(ic), 
   read(*addMasterPort("read", uint8_pinout)),
   poll(*addMasterPort("poll", uint8_pinout)),
   safe(s),
   ready(false),
   report(0),
   polling(false) {
}

SendQueue::const_iterator LM32Read::send(SendQueue::const_iterator i, SendQueue::const_iterator end) {
  for (; i != end; ++i) {
    if (i->master == &reset) { 
      ready = false;
      polling = false;
      report = 0;
    } else if (i->master == &poll) {
      if (!ready && !polling && !report) {
        polling = true;
        read_byte();
      }
    } else {
      /* Block until read is ready */
      if (!ready) {
        if (!polling && !report) {
          polling = true;
          read_byte();
        }
        return i;
      }
      
      /* We only would poll if not ready. */
      assert (!polling);
      
      send_keyval(UART_READ, 0);
      ++report;
      
      if (!safe) {
        /* We pipeline consequetive reads into each other */
        while (++i != end && i->master == &read) {
          sendrecv_keyval(UART_READ, 0);
          ++report;
        }
        --i;
      }
      
      read_byte(); /* grab the (last) result */
      ready = false;
    }
  }
  
  return i;
}

void LM32Read::recv(unsigned int status, uint8_t value) {
  if (report) {
    read(Bits(value));
    --report;
  } else {
    assert(polling);
  }
  
  ready = (status & UART_READ_READY) != 0;
  polling = false;
  
  if (!ready && report) 
    throw ProtocolViolation(this, "LM32 device driver too slow; use safe mode for reads");
}

bool LM32Read::amReady() const {
  return ready;
}
